from textwrap import dedent

ORIGINAL = dedent(
    '''\
    original first line
    original second line
    original third line
'''
)

CHANGED = dedent(
    '''\
    original first line
    changed second line
    original third line
'''
)
